#ifndef INCLUDE_GUARD_1648B36D_0575_4B12_9CB6_B82467427338
#define INCLUDE_GUARD_1648B36D_0575_4B12_9CB6_B82467427338

#include <Python.h>

namespace cas {

extern PyObject* enum_type;
extern PyObject* enum_status;
extern PyObject* enum_severity;

extern PyObject* enum_exists;
extern PyObject* enum_attach;

}

#endif
