#ifndef INCLUDE_GUARD_7B46E958_448E_4FE7_9DFD_75E5267FB333
#define INCLUDE_GUARD_7B46E958_448E_4FE7_9DFD_75E5267FB333

#define PY_ARRAY_UNIQUE_SYMBOL CA_SERVER_ARRAY_API
#define NPY_NO_DEPRECATED_API NPY_1_8_API_VERSION
#include <numpy/arrayobject.h>

#endif
